## Styles

Everything is immutable in ScalaCSS
(except for [`mutable.Register`](https://github.com/japgolly/scalacss/blob/master/core/src/main/scala/japgolly/scalacss/mutable/Register.scala)
which collects and remembers styles as you define them),
so any `StyleS`,
[`StyleF`](stylef.md),
[`StyleC`](nested.md)
can safely be used anywhere else.

`StyleA` (_applicable_ inline style) is immutable too but one of its properties
is a generated class name. If you're using inline stylesheet, class names are
generated by [`mutable.Register`](https://github.com/japgolly/scalacss/blob/master/core/src/main/scala/japgolly/scalacss/mutable/Register.scala)
so sharing them is not a good idea. Instead what you want is a reusable style module.

## Style Modules

It's often preferable to share a collection of named styles, a style _module_.

The definition of a style module is only slighty different from normal stylesheets
in that instead of defining an `object`
```scala
object MyStyles extends StyleSheet.Inline {
```
you create a class with an implicit register
```scala
class MyStyleModule(implicit r: mutable.Register) extends StyleSheet.Inline()(r) {
```

_Note: The `()(r)` syntax is weird because Scala has a rule that implicit arguments
always come second, even when there are no non-implicit arguments. That is the case
here which is why it forces us to specify an empty argument list before setting the
implicit one._

To use shared styles, you simply create an instance of them inside another stylesheet.

```scala
object MyStyles extends StyleSheet.Inline {
  val shared = new MyStyleModule
}
```

That's all there is to it.

## Example

A shared module:
```scala
class SharedTheme(implicit r: mutable.Register) extends StyleSheet.Inline()(r) {
  import dsl._

  val button = style(
    padding(0.5 ex, 2 ex),
    backgroundColor("#eee"),
    border(1 px, solid, black)
  )

  val title = style(
    fontSize(32 px)
  )
}
```

Concrete styles in the app:
```scala
object MyAppStyles extends StyleSheet.Inline {
  import dsl._

  val theme = new SharedTheme // Instantiate shared styles

  val headingTitle = style(
    theme.title,              // Extend shared style
    color(red)
  )
}
```

Example usage in [scalajs-react](https://github.com/japgolly/scalajs-react):
```scala
def render =
  <.div(
    <.h1(
      MyAppStyles.headingTitle,
      "Your Login Details"),
    <.button(
      MyAppStyles.theme.button,
      ^.onClick ~~> showLoginDetails(user),
      "Show Details"))
```
